import multiprocessing

class Semaps:
	def __init__ (self, keys, type = "d", slots = 256):
		self.slots = slots
		self._keys = []
		if type == "d":
			initial_val = 0.0
		else:
			initial_val = 0
		self._arr = multiprocessing.Array (type, [initial_val] * self.slots, lock = multiprocessing.RLock ())
		self._d = {}
		self.add (keys)

	def add (self, keys):
		initial = len (self._keys)
		for i, key in enumerate (keys):
			assert isinstance (key, str)
			if key in self._d:
				continue
			self._d [key] = initial + i
			self._keys.append (key)

	def __len__ (self):
		return len (self._d)

	def __contains__ (self, k):
		return k in self._d

	def __setitem__ (self, k, v):
		self.set (k, v)

	def __getitem__ (self, k):
		return self._d [k]

	def set (self, k, v, ignore_nokey = False):
		try:
			self._arr [self._d [k]] = v
		except KeyError:
			if not ignore_nokey:
				raise

	def get (self, k, d = None):
		v = self._arr [self._d [k]]
		if not v and d:
			self._arr [self._d [k]] = d
			return d
		return v

	def has_key (self, k):
		return self._d.has_key (k)

	def keys (self):
		return self._keys [:]

	def values (self):
		t = []
		for i in range (len (self._keys)):
			t.append (self._arr [i])
		return t

	def items (self):
		t = []
		for i in range (len (self._keys)):
			t.append ((self._keys [i], self._arr [i]))
		return t


class TestSemaps (Semaps):
    def __init__ (self, keys = [], tpye= "d", slots = 256):
        self._keys = keys
        self._arr = [0] * slots
        self._d = {}

    def set (self, k, v, ignore_nokey = False):
        if k not in self._d:
            self._d [k] = len (self._d)
        self._arr [self._d [k]] = v



import os, sys, re, types, time
from rs4  import pathtool, importer, evbus
from rs4.termcolor import tc
import threading
from types import FunctionType as function
import copy
from .. import lifetime
import inspect
from rs4 import attrdict
from importlib import reload

def set_default (cf):
    cf.max_post_body_size = 1 * 1024 * 1024
    cf.max_cache_size = 1 * 1024 * 1024
    cf.max_multipart_body_size = 20 * 1024 * 1024
    cf.max_upload_file_size = 20 * 1024 * 1024

def Config (preset = False):
    cf = attrdict.AttrDict ()
    if preset:
        set_default (cf)
    return cf


class DjangoReloader:
    def __init__ (self, mounted, logger):
        from django.utils import autoreload

        self.mounted = mounted
        self.logger = logger
        self.mtimes = {}
        self.version = 1
        if hasattr (autoreload, "code_changed"):
            self.reloader = autoreload
        else:
            self.version = 2
            self.reloader = autoreload.get_reloader ()

    def reloaded (self):
        if self.version == 1:
            return self.reloader.code_changed ()
        else:
            for filepath, mtime in self.reloader.snapshot_files ():
                if not str (filepath).startswith (self.mounted):
                    continue
                old_time = self.mtimes.get (filepath)
                self.mtimes [filepath] = mtime
                if old_time is None:
                    continue
                elif mtime > old_time:
                    return True
        return False

class Module:
    def __init__ (self, wasc, handler, bus, route, directory, libpath, pref = None):
        self.wasc = wasc
        self.bus = bus
        self.handler = handler
        self.pref = pref
        self.last_reloaded = time.time ()
        self.set_route (route)
        self.directory = directory
        self.has_life_cycle = False

        self.app = None
        self.django = False
        self.debug = False
        self.use_reloader = False

        if type (libpath) is str:
            try:
                libpath, self.appname = libpath.split (":", 1)
            except ValueError:
                libpath, self.appname = libpath, "app"
            self.libpath = libpath
            self.script_name = "%s.py" % libpath
            self.module, self.abspath = importer.importer (directory, libpath)
            self.start_app ()

        else:
            # libpath is app object, might be added by unittest
            self.appname, self.module, self.abspath = "app", None, os.path.join (directory, 'dummy')
            self.app = libpath
            self.app.use_reloader = False
            self.start_app ()

        app = self.app or getattr (self.module, self.appname)
        self.has_life_cycle and app.life_cycle ("before_mount", self.wasc)

    def __repr__ (self):
        return "<Module routing to %s at %x>" % (self.route, id(self))

    def get_callable (self):
        return self.app or getattr (self.module, self.appname)

    def start_app (self, reloded = False):
        func = None
        app = self.app or getattr (self.module, self.appname)

        if hasattr (app, "set_logger"):
            app.set_logger (self.wasc.logger.get ("app"))

        self.django = str (app.__class__).find ("django.") != -1 and DjangoReloader (os.path.dirname (os.path.dirname (self.abspath)), self.wasc.logger)
        self.has_life_cycle = hasattr (app, "life_cycle")

        if self.pref:
            for k, v in copy.copy (self.pref).items ():
                if k == "config":
                    if not hasattr (app, 'config'):
                        app.config = v
                    else:
                        for k, v in copy.copy (self.pref.config).items ():
                            app.config [k] = v
                else:
                    setattr (app, k, v)

        if hasattr (app, "_aliases"):
            while app._aliases:
                self.wasc.add_cluster (*app._aliases.pop (0))

        if not hasattr (app, "config"):
            app.config = Config (True)

        elif not hasattr (app.config, "max_post_body_size"):
            set_default (app.config)

        if hasattr (app, "mountables"):
            for _args, _karg in app.mountables:
                app.mount (*_args, **_karg)

        if hasattr (app, "max_client_body_size"):
            val = app.max_client_body_size
            app.config.max_post_body_size = val
            app.config.max_multipart_body_size = val
            app.config.max_upload_file_size = val

        if hasattr (app, "set_home"):
            app.set_home (os.path.dirname (self.abspath), self.module)

        if hasattr (app, "commit_events_to"):
            app.commit_events_to (self.bus)

        self.set_devel_env ()
        self.update_file_info ()

        try:
            if not reloded:
                func = app.start
            else:
                func = app.restart
        except AttributeError:
            pass

        if func:
            # temporary current handler
            self.wasc.handler = self.handler
            func (self.wasc, self.route)
            self.wasc.handler = None

    def mounted (self):
        app = self.app or getattr (self.module, self.appname)
        self.has_life_cycle and app.life_cycle ("mounted", self.wasc ())
        self.has_life_cycle and app.life_cycle ("mounted_or_reloaded", self.wasc ())

    def umounted (self):
        app = self.app or getattr (self.module, self.appname)
        self.has_life_cycle and app.life_cycle ("umounted", self.wasc)

    def cleanup (self):
        app = self.app or getattr (self.module, self.appname)
        self.has_life_cycle and app.life_cycle ("before_umount", self.wasc ())
        try: app.cleanup ()
        except AttributeError: pass

    def check_django_reloader (self, now):
        if self.django.reloaded ():
            self.wasc.logger ("app", "reloading app, %s" % self.abspath, "debug")
            self.last_reloaded = time.time ()
            lifetime.shutdown (3, 0)

    def set_devel_env (self):
        app = self.app or getattr (self.module, self.appname)
        skitai_env = os.environ.get ("SKITAIENV")
        if skitai_env == "DEVEL":
            self.debug = app.debug = True
            self.use_reloader = app.use_reloader = True
        elif skitai_env == "SILENT":
            self.debug = app.debug = False
            self.use_reloader = app.use_reloader = False
        else:
            # inherit
            try: self.debug = app.debug
            except AttributeError: pass
            try: self.use_reloader = app.use_reloader
            except AttributeError: pass

        if self.use_reloader and self.django:
            lifetime.maintern.sched (1.0, self.check_django_reloader)

    def update_file_info (self):
        if self.module is None:
            # app directly mounted
            return
        stat = os.stat (self.abspath)
        self.file_info = (stat.st_mtime, stat.st_size)

    def maybe_reload (self):
        if not self.use_reloader:
            return

        if self.django:
            # see check_django_reloader
            return

        if time.time () - self.last_reloaded < 1.0:
            return

        stat = os.stat (self.abspath)
        if self.file_info != (stat.st_mtime, stat.st_size):
            oldapp = getattr (self.module, self.appname)
            self.has_life_cycle and oldapp.life_cycle ("before_reload", self.wasc ())

            try:
                reloaded = importer.reimporter (self.module, self.directory, self.libpath)
            except:
                self.module.app = oldapp
                raise
            else:
                if not reloaded:
                    return
                self.module, self.abspath = reloaded
                if hasattr (oldapp, "remove_events"):
                    oldapp.remove_events (self.bus)
                PRESERVED = []
                if hasattr (oldapp, "PRESERVES_ON_RELOAD"):
                    PRESERVED = [(attr, getattr (oldapp, attr)) for attr in oldapp.PRESERVES_ON_RELOAD]

                self.start_app (reloded = True)
                newapp = getattr (self.module, self.appname)
                for attr, value in PRESERVED:
                    setattr (newapp, attr, value)

                # reloaded
                self.has_life_cycle and newapp.life_cycle ("reloaded", self.wasc ())
                self.has_life_cycle and newapp.life_cycle ("mounted_or_reloaded", self.wasc ())
                self.last_reloaded = time.time ()
                self.wasc.logger ("app", "reloading app, %s" % self.abspath, "debug")

    def set_route (self, route):
        route = route
        while route and route [-1] == "/":
            route = route [:-1]
        self.route = route + "/"
        self.route_len = len (self.route) - 1

    def get_route (self):
        return self.route

    def get_path_info (self, path):
        path_info = ("/" + path) [self.route_len:]
        return path_info

    def __call__ (self, env, start_response):
        self.use_reloader and self.maybe_reload ()
        app = self.app or getattr (self.module, self.appname)
        return app (env, start_response)


class ModuleManager:
    modules = {}
    def __init__(self, wasc, handler):
        self.wasc = wasc
        self.handler = handler
        self.modules = {}
        self.modnames = {}
        self.bus = evbus.EventBus ()
        self.cc = 0

    def __getitem__ (self, name):
        return self.modnames [name].get_callable ()

    def build_url (self, thing, *args, **kargs):
        a, b = thing.split (":", 1)
        return self.modnames [a].get_callable ().build_url (b, *args, **kargs)

    def add_module (self, route, directory, modname, pref):
        dispname = modname
        if type (modname) is str and  modname.startswith ("__export__"):
            dispname = "{}:app".format (os.path.basename (os.path.dirname (os.path.dirname (directory))))

        if dispname in self.modnames:
            self.wasc.logger ("app", "app file name collision detected '%s'" % dispname, "error")
            return

        if not route:
            route = "/"
        elif not route.endswith ("/"):
            route = route + "/"

        try:
            module = Module (self.wasc, self.handler, self.bus, route, directory, modname, pref)
        except:
            self.wasc.logger.trace ("app")
            self.wasc.logger ("app", "[error] app load failed: %s" % dispname)
        else:
            self.wasc.logger ("app", "[info] app %s mounted to %s" % (dispname, tc.white (route)))
            if route in self.modules:
                self.wasc.logger ("app", "[info] app route collision detected: %s at %s <-> %s" % (route, module.abspath, self.modules [route].abspath), "warn")
            self.modules [route] = module
            if type (dispname) is str:
                # possibley direct app object
                self.modnames [dispname.split (":", 1)[0]] = module

    def get_app (self, script_name):
        route = self.has_route (script_name)
        if route in (0, 1): # 404, 301
            return None

        try:
            app = self.modules [route]
        except KeyError:
            return None
        return app

    def has_route (self, script_name):
        # 0: 404
        # 1: 301 => /skitai => /skitai/
        if script_name == "/" and "/" in self.modules:
            return "/"

        cands = []
        for route in self.modules:
            if script_name == route:
                cands.append (route)
            elif script_name + "/" == route:
                return 1
            elif script_name.startswith (route [-1] != "/" and route + "/" or route):
                cands.append (route)

        if cands:
            if len (cands) == 1:
                return cands [0]
            cands.sort (key = lambda x: len (x))
            return cands [-1]

        elif "/" in self.modules:
            return "/"

        return 0

    def unload (self, route):
        module = self.modules [route]
        self.wasc.logger ("app", "[info] unloading app: %s" % route)
        try:
            self.wasc.logger ("app", "[info] ..cleanup app: %s" % route)
            module.cleanup ()
        except AttributeError:
            pass
        except:
            self.wasc.logger.trace ("app")
        del module
        del self.modules [route]

    def cleanup (self):
        self.wasc.logger ("app", "[info] cleanup apps")
        for route, module in list(self.modules.items ()):
            try:
                self.wasc.logger ("app", "[info] ..cleanup app: %s" % route)
                module.cleanup ()
            except AttributeError:
                pass
            except:
                self.wasc.logger.trace ("app")

    def status (self):
        d = {}
        for path, module in list(self.modules.items ()):
            d ['<a href="%s">%s</a>' % (path, path)] = module.abspath
        return d





from aquests.protocols.http import request, response
from aquests.protocols.grpc import request as grpc_request
from aquests.protocols.ws import request as ws_request
from aquests.dbapi import request as dbo_request
import skitai
from ..backbone.http_request import http_request
from base64 import b64encode
import os
from . import channel
from .server import get_client_response
import json
from rs4 import siesta
from urllib.parse import urlparse

DEFAULT_HEADERS = {
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8",
    "Accept-Encoding": "gzip, deflate, br",
    "Accept-Language": "en-US,en;q=0.8,en-US;q=0.6,en;q=0.4",
    "Referer": "https://pypi.python.org/pypi/skitai",
    "Upgrade-Insecure-Requests": 1,
    "User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/59.0.3071.115 Safari/537.36",
    "Host": "skitai.com"
}

class Method (siesta.Resource):
    ACCPET = "*/*"
    def __call__(self, *args):
        script = urlparse (self._api.base_url) [2]
        method = self._url [len (script):]
        return self._callback (self._api.base_url, method.replace ("/", "."), args)

class RPC (siesta.API):
    RESOURCE_CLASS = Method
    def __call__(self):
        raise AttributeError

class Client:
    def override (self, headers):
        copied = DEFAULT_HEADERS.copy ()
        for k, v in headers:
            copied [k] = v
        return copied

    def handle_request (self, request, handler = None):
        # clinet request -> process -> server response -> client response
        return get_client_response (request, handler)

    def __serialize (self, payload):
        _payload = []
        while 1:
            d = payload.more ()
            if not d: break
            _payload.append (d)
        return b"".join (_payload)

    def __generate (self, r):
        m, u, v = r.get_method (), r.path, r.http_version
        headers = ["%s: %s" % each for each in r.get_headers ()]
        return http_request (channel.Channel (), "%s %s HTTP/%s" % (m, u, v), m.lower (), u, v, headers)

    def make_request (self, method, uri, data, headers, auth = None, meta = {}, version = "1.1"):
        method = method.upper ()
        if isinstance (headers, dict):
            headers = [(k, v) for k, v in headers.items ()]
        headers = self.override (headers)
        if method == "UPLOAD":
            r = request.HTTPMultipartRequest (uri, "POST", data, headers, auth, None, meta, version)
        else:
            r = request.HTTPRequest (uri, method, data, headers, auth, None, meta, version)
        hr = self.__generate (r)
        if data:
            payload = r.get_payload ()
            if method == "UPLOAD":
                payload = self.__serialize (payload)
            hr.set_body (payload)
        return hr

    def get (self, uri, headers = [], auth = None, meta = {}, version = "1.1"):
        return self.make_request ("GET", uri, None, headers, auth, meta, version)

    def delete (self, uri, headers = [], auth = None, meta = {}, version = "1.1"):
        return self.make_request ("DELETE", uri, None, headers, auth, meta, version)

    def post (self, uri, data, headers = [], auth = None, meta = {}, version = "1.1"):
        return self.make_request ("POST", uri, data, headers, auth, meta, version)

    def postjson (self, uri, data, headers = [], auth = None, meta = {}, version = "1.1"):
        headers.append (('Accpet', 'application/json'))
        headers.append (('Content-Type', 'application/json'))
        return self.make_request ("POST", uri, data, headers, auth, meta, version)

    def patch (self, uri, data, headers = [], auth = None, meta = {}, version = "1.1"):
        return self.make_request ("PATCH", uri, data, headers, auth, meta, version)

    def put (self, uri, data, headers = [], auth = None, meta = {}, version = "1.1"):
        return self.make_request ("PUT", uri, data, headers, auth, meta, version)

    def upload (self, uri, data, headers = [], auth = None, meta = {}, version = "1.1"):
        return self.make_request ("UPLOAD", uri, data, headers, auth, meta, version)

    # api ----------------------------------------------------
    def api (self, endpoint = ""):
        return siesta.API (endpoint, callback = self.make_request)

    # rpc ----------------------------------------------------
    def rpc (self, endpoint = ""):
        return RPC (endpoint, callback = self.__continue_xmlrpc)
    xmlrpc = rpc

    def __continue_xmlrpc (self, uri, method, data, headers = [], auth = None, meta = {}, version = "1.1"):
        r = request.XMLRPCRequest (uri, method, data, self.override (headers), auth, None, meta, version)
        hr = self.__generate (r)
        hr.set_body (r.get_payload ())
        hr._xmlrpc_serialized = True
        return hr

    def jsonrpc (self, endpoint = ""):
        return RPC (endpoint, callback = self.__continue_jsonrpc)

    def __continue_jsonrpc (self, uri, method, data, headers = [], auth = None, meta = {}, version = "1.1"):
        r = request.JSONRPCRequest (uri, method, data, self.override (headers), auth, None, meta, version)
        hr = self.__generate (r)
        hr.set_body (r.get_payload ())
        return hr

    def grpc (self, endpoint = ""):
        return RPC (endpoint, callback = self.__continue_grpc)

    def __continue_grpc (self, uri, method, data, headers = [], auth = None, meta = {}, version = "2.0"):
        r = grpc_request.GRPCRequest (uri, method, data, self.override (headers), auth, None, meta, version)
        hr = self.__generate (r)
        hr.set_body (self.__serialize (r.get_payload ()))
        return hr

    # db ----------------------------------------------------
    def __make_dbo (self, dbtype, server, dbname = None, auth = None, method = None, params = None, callback = None, meta = {}):
        return dbo_request.Request (dbtype, server, dbname, auth, method, params, callback, meta)

    def postgresql (self, server, dbname, *args, **kargs):
        return self.__make_dbo ('postgresql', server, dbname, *args, **kargs)

    def mongdb (self, serverr, dbname, *args, **kargs):
        return self.__make_dbo ('mongodb', serverr, dbname, *args, **kargs)

    def sqlite3 (self, server, *args, **kargs):
        return self.__make_dbo ('sqlite3', server, *args, **kargs)

    def redis (self, server, *args, **kargs):
        return self.__make_dbo ('redis', server, *args, **kargs)

    # websokcet --------------------------------------------
    def ws (self, uri, message, headers = [], auth = None, meta = {}, version = "1.1"):
        r = ws_request.Request (uri, message, self.override (headers), auth, None, meta, version)
        origin = r.address or ("127.0.0.1", 80)
        r.headers ['Origin'] = "http://%s:%d" % origin
        r.headers ['Sec-WebSocket-Key'] = b64encode(os.urandom(16))
        r.headers ['Connection'] = "keep-alive, Upgrade"
        r.headers ['Upgrade'] = 'websocket'
        r.headers ['Cache-Control'] = 'no-cache'
        r.method = "get"
        hr = self.__generate (r)
        hr.set_body (r.get_payload ().more ())
        return self.handle_request (hr)



from ...backbone import http_response
from rs4 import producers
import time

class response (http_response.http_response):
	def __init__ (self, request):
		http_response.http_response.__init__ (self, request)
		self.trailers = []

	def set_trailer (self, name, value):
		self.trailers.append ((name, value))

	def get_trailers (self):
		return self.trailers

	def set_streaming (self):
		self._is_async_streaming = True

	def build_reply_header (self):
		h = [(b":status", str (self.reply_code).encode ("utf8"))]
		for k, v in self.reply_headers:
			h.append ((k.lower ().encode ("utf8"), str (v).encode ("utf8")))
		return h

	def push_promise (self, uri):
		if not self.request.protocol.pushable ():
			return

		host = self.request.get_header ('host', '')
		headers = [
			(':path', uri),
			(':authority', host),
			(':scheme', self.request.scheme),
			(':method', "GET")
	  	]
		additional_headers = [
	   		(k, v) for k, v in [
			   	('accept-encoding', self.request.get_header ('accept-encoding')),
			   	('accept-language', self.request.get_header ('accept-language')),
			    ('cookie', self.request.get_header ('cookie')),
			    ('user-agent', self.request.get_header ('user-agent')),
			    ('referer', "%s://%s%s" % (self.request.scheme, host, self.request.uri)),
		   ] if v
	   ]
		self.request.protocol.push_promise (self.request.stream_id, headers, additional_headers)
	hint_promise = push_promise

	def done (self, force_close = False, upgrade_to = None):
		self.content_type = self.get ('content-type')

		if not self.is_responsable (): return
		self._is_done = True
		if self.request.protocol is None: return

		self.htime = (time.time () - self.stime) * 1000
		self.stime = time.time () #for delivery time

		# removed by HTTP/2.0 Spec.
		self.delete ('transfer-encoding')
		self.delete ('connection')

		# compress payload and globbing production
		do_optimize = True
		if upgrade_to or self.is_async_streaming ():
			do_optimize = False

		if not self.outgoing:
			self.delete ('content-type')
			self.delete ('content-length')
			outgoing_producer = None

		elif len (self.outgoing) == 1 and hasattr (self.outgoing.first (), "ready"):
			outgoing_producer = producers.composite_producer (self.outgoing)
			do_optimize = False

		else:
			outgoing_producer = producers.composite_producer (self.outgoing)
			if do_optimize and not self.has_key ('Content-Encoding'):
				way_to_compress = ""
				maybe_compress = self.request.get_header ("Accept-Encoding")

				if maybe_compress:
					cl = self.has_key ("content-length") and int (self.get ("Content-Length")) or -1
					if cl == -1:
						cl = self.outgoing.get_estimate_content_length ()
					if 0 < cl <= http_response.UNCOMPRESS_MAX:
						maybe_compress = ""

				if maybe_compress:
					content_type = self.get ("Content-Type")
					if maybe_compress and content_type and (content_type.startswith ("text/") or content_type.startswith ("application/json")):
						accept_encoding = [x.strip () for x in maybe_compress.split (",")]
						if "gzip" in accept_encoding:
							way_to_compress = "gzip"
						elif "deflate" in accept_encoding:
							way_to_compress = "deflate"

				if way_to_compress:
					if self.has_key ('Content-Length'):
						self.delete ("content-length") # rebuild
					self.update ('Content-Encoding', way_to_compress)
					if way_to_compress == "gzip":
						compressing_producer = producers.gzipped_producer
					else: # deflate
						compressing_producer = producers.compressed_producer
					outgoing_producer = compressing_producer (outgoing_producer)

		if self.request.protocol is None:
			return

		if upgrade_to:
			# do not change http2 channel
			request, terminator = upgrade_to
			self.request.channel.current_request = request
			self.request.channel.set_terminator (terminator)

		logger = self.request.logger #IMP: for disconnect with request
		try:
			self.request.protocol.handle_response (
				self.request.stream_id,
				self.build_reply_header (),
				self.get_trailers (),
				outgoing_producer,
				do_optimize,
				force_close = force_close
			)
		except:
			logger.trace ()



import pstats
import os
from rs4.termcolor import tc
import getopt, sys

def main ():
    if not os.path.isfile ('profile.out'):
        print (tc.error ("error:"))
        print (" - profile.out file not found")
        print (" - for creating this file, run your app with ---profile")
        print ()
        return

    argopt = getopt.getopt (sys.argv[1:], "n:s:", ["sort="])
    SORT = 'ncalls'
    N = 300
    for k, v in argopt [0]:
        if k == '-s' or k == '--sort':
            SORT = v
        elif k == '-n':
            N = int (v)

    stats = pstats.Stats ('profile.out')
    stats.strip_dirs()
    stats.sort_stats(SORT)
    stats.print_stats(N)
    stats.print_callers ()




from . import http2_handler
import time
import os
from aioquic.quic import events
from aioquic.h3.exceptions import H3Error, NoAvailablePushIDError
from aioquic.buffer import Buffer
from aioquic.quic.packet import PACKET_TYPE_INITIAL, encode_quic_retry, encode_quic_version_negotiation, pull_quic_header
from aioquic.quic.retry import QuicRetryTokenHandler
from aioquic.quic.connection import QuicConnection
import enum
from aquests.protocols.http3.events import PushCanceled, MaxPushIdReceived, DataReceived, HeadersReceived
from aquests.protocols.http3.connection import H3Connection

# http2 compat error codes
class ErrorCode (enum.IntEnum):
    NO_ERROR = 0x0
    PROTOCOL_ERROR = 0xA
    INTERNAL_ERROR = 0x1
    FLOW_CONTROL_ERROR = 0x3
    CANCEL = 0x0

class http3_producer (http2_handler.http2_producer):
    def local_flow_control_window (self):
        return self.SIZE_BUFFER

class http3_request_handler (http2_handler.http2_request_handler):
    producer_class = http3_producer
    stateless_retry = True
    conns = {}
    errno = ErrorCode

    def __init__ (self, handler, request):
        self._default_varialbes (handler, request)
        self.quic = None # QUIC protocol
        self.conn = None # HTTP3 Protocol

        self._push_map = {}
        self._retry = QuicRetryTokenHandler() if self.stateless_retry else None

    def _initiate_shutdown (self):
        # pahse I: send goaway
        with self._plock:
            self.conn.shutdown (self._shutdown_reason [-1])

    def _proceed_shutdown (self):
        # phase II: close quic
        if self.quic:
            errcode, msg, _ = self._shutdown_reason
            with self._plock:
                self.quic.close (errcode, reason_phrase = msg or '')
            self.send_data ()

    def _make_connection (self, channel, data):
        # https://github.com/aiortc/aioquic/commits/master/src/aioquic/asyncio/server.py
        ctx = channel.server.ctx

        buf = Buffer (data=data)
        header = pull_quic_header (
            buf, host_cid_length = ctx.connection_id_length
        )
        # version negotiation
        if header.version is not None and header.version not in ctx.supported_versions:
            self.channel.push (
                encode_quic_version_negotiation (
                    source_cid = header.destination_cid,
                    destination_cid = header.source_cid,
                    supported_versions = ctx.supported_versions,
                )
            )
            return

        conn = self.conns.get (header.destination_cid)
        if conn:
            conn._linked_channel.close ()
            conn._linked_channel = channel
            self.quic = conn._quic
            self.conn = conn
            return

        if header.packet_type != PACKET_TYPE_INITIAL or len (data) < 1200:
            return

        original_destination_connection_id = None
        retry_source_connection_id = None
        if self._retry is not None:
            source_cid = os.urandom(8)
            if not header.token:
                # create a retry token
                channel.push (
                    encode_quic_retry (
                        version = header.version,
                        source_cid = source_cid,
                        destination_cid = header.source_cid,
                        original_destination_cid = header.destination_cid,
                        retry_token = self._retry.create_token (channel.addr, header.destination_cid, source_cid)))
                return
            else:
                try:
                    original_destination_connection_id, retry_source_connection_id = self._retry.validate_token (channel.addr, header.token)
                except ValueError:
                    return
        else:
            original_destination_connection_id = header.destination_cid

        self.quic = QuicConnection (
            configuration = ctx,
            original_destination_connection_id = original_destination_connection_id,
            retry_source_connection_id = retry_source_connection_id,
            session_ticket_fetcher = channel.server.ticket_store.pop,
            session_ticket_handler = channel.server.ticket_store.add
        )
        self.conn = H3Connection (self.quic)
        self.conn._linked_channel = channel

    def _handle_events (self, events):
        for event in events:
            if isinstance (event, HeadersReceived):
                created = self.handle_request (event.stream_id, event.headers, has_data_frame = not event.stream_ended)
                if created:
                    r = self.get_request (event.stream_id)
                    if event.stream_ended:
                        r.set_stream_ended ()

            elif isinstance (event, DataReceived) and event.data:
                r = self.get_request (event.stream_id)
                if not r:
                    self.close (self.errno.INTERNAL_ERROR)
                else:
                    try:
                        r.channel.set_data (event.data, len (event.data))
                    except ValueError:
                        self.close (self.errno.INTERNAL_ERROR)

                    if event.stream_ended:
                        if r.collector:
                            r.channel.handle_read ()
                            r.channel.found_terminator ()
                        r.set_stream_ended ()
                        if r.response.is_done ():
                            self.remove_request (event.stream_id)

            elif isinstance (event, PushCanceled):
                with self._clock:
                    try: del self._push_map [event.push_id]
                    except KeyError: pass
                self.remove_push_stream (event.push_id)

        self.send_data ()

    def process_quic_events (self):
        while 1:
            with self._plock:
                event = self.quic.next_event ()
            if event is None:
                break

            if isinstance (event, events.StreamDataReceived):
                h3_events = self.conn.handle_event (event)
                h3_events and self._handle_events (h3_events)

            elif isinstance (event, events.ConnectionIdIssued):
                self.conns [event.connection_id] = self.conn

            elif isinstance (event, events.ConnectionIdRetired):
                assert self.conns [event.connection_id] == self.conn
                conn = self.conns.pop (event.connection_id)
                conn._linked_channel = None

            elif isinstance (event, events.ConnectionTerminated):
                for cid, conn in list (self.conns.items()):
                    if conn == self.conn:
                        conn._linked_channel = None
                        del self.conns [cid]
                self._terminate_connection ()

            elif isinstance (event, events.HandshakeCompleted):
                pass

            elif isinstance (event, events.PingAcknowledged):
                # for now nothing to do, channel will be extened by event time
                pass

    def collect_incoming_data (self, data):
        # print ('collect_incoming_data', self.quic, len (data))
        if self.quic is None:
             self._make_connection (self.channel, data)
             if self.quic is None:
                 return
        with self._plock:
            self.quic.receive_datagram (data, self.channel.addr, time.monotonic ())
        self.process_quic_events ()
        self.send_data ()

    def reset_stream (self, stream_id):
        raise AttributeError ('HTTP/3 can cancel for only push stream, use cancel_push (push_id)')

    def remove_stream (self, stream_id):
        raise AttributeError ('use remove_push_stream (push_id)')

    def remove_push_stream (self, push_id):
        # received by client and just reset
        with self._clock:
            try: stream_id = self._push_map.pop (push_id)
            except KeyError: return
        super ().remove_stream (stream_id)

    def cancel_push (self, push_id):
        # send to client and reset
        with self._clock:
            try: steram_id = self._push_map.pop (push_id)
            except KeyError: return # already done or canceled
        with self._plock:
            self.conn.cancel_push (push_id)
        super ().remove_stream (steram_id)
        self.send_data ()

    def data_to_send (self):
        if self.quic is None or self._closed:
            return []
        self._data_from_producers ()
        with self._plock:
            data_to_send = [data for data, addr in self.quic.datagrams_to_send (now = time.monotonic ())]
        return data_to_send or self._data_exhausted ()

    def pushable (self):
        return self.request_acceptable ()

    def push_promise (self, stream_id, request_headers, addtional_request_headers):
        name_, path = request_headers [0]
        assert name_ == ':path', ':path header missing'
        headers = [(k.encode (), v.encode ()) for k, v in request_headers + addtional_request_headers]
        try:
            promise_stream_id = self.conn.send_push_promise (stream_id = stream_id, headers = headers)
        except NoAvailablePushIDError:
            return

        with self._clock:
            push_id = self._pushed_pathes.get (path)

        if push_id is None:
            try:
                push_id = self.conn.get_push_id (promise_stream_id)
            except AttributeError:
                push_id = self.conn._next_push_id - 1

            with self._clock:
                self._push_map [push_id] = promise_stream_id
                self._pushed_pathes [path] = push_id

        self._handle_events ([HeadersReceived (headers = headers, stream_ended = True, stream_id = promise_stream_id, push_id = push_id)])


class Handler (http2_handler.Handler):
    # keep_alive = 37 same as http2
    # chrome QUIC timout is 30s but first time it extends more 30s only one time
    def match (self, request):
        return request.version.startswith ("3.")

    def handle_request (self, request):
        http3 = http3_request_handler (self, request)
        request.channel.die_with (http3, "http3 connection")
        request.channel.set_socket_timeout (self.keep_alive)
        request.channel.current_request = http3
